module background_mod

! this is for omgrism-1.9
! This module generates the source list detected in the input image
!

!
   use dal
   use sasdal
   use errorHandling
   use OmParameters_mod ! for computing the source moments
   use omconstants_mod

! VNY 16.12.2004
   use OmSourceParam_mod ! Source definitions

   implicit none

   private

!   public SimpleImageSmoothing
   public FindCentralEnhancement

contains





!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! CheckConnectedPixels
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
recursive function  CheckConnectedPixels &
     (image, bkgImage, imaskImage, xMin, xMax, yMin, yMax, i,j, label, &
     sumX, sumY, sumValue, bkgValue, verbFlag, currentNpix) result(nPixels)
  integer(int32), intent(in) :: xMin, xMax, yMin, yMax
  real(single), intent(in) :: image(xMin:xMax, yMin:yMax)
  real(single), intent(in) :: bkgImage(xMin:xMax, yMin:yMax)
  integer(int32), intent(inout) :: imaskImage(xMin:xMax, yMin:yMax)       
  integer(int32), intent(in) :: i,j
  integer(int32), intent(in) :: label
  real(single), intent(inout) :: sumX, sumY, sumValue, bkgValue
  integer(int32), intent(inout) :: currentNpix
  integer(int32) :: k,l
  integer(int32) :: nPixels
  integer(int32) :: nPixels0, nPixels1
  integer(int32) :: verbFlag
  nPixels0=0
  if (currentNpix .gt. 10000) then
!  if (currentNpix .gt. 30000) then
     nPixels=0
     return
  endif

  do k=i-1,i+1
     do l=j-1,j+1
        if (k .ge. xMin .and. k .le. xMax &
             .and. l .ge. yMin .and. l .le.yMax) then
           if (imaskImage(k,l) .gt. 0) then
              nPixels0=nPixels0+1
              currentNpix=currentNpix+1

              imaskImage(k,l)=-label
              !if (verbFlag .eq. 1) write(*,*) "L1714 k=", k, " l=", l
                 if (image(k,l) .gt. sumValue) then
                    sumValue=image(k,l)
                    bkgValue=bkgImage(k,l)
                    sumX=real(k)
                    sumY=real(l)
                 endif
                 !write(*,*) " L1825"
              nPixels1=CheckConnectedPixels(image, bkgImage, &
                   imaskImage,xMin,xMax,yMin,yMax,k,&
                   l,label, sumX, sumY, sumValue, bkgValue, verbFlag, &
                   currentNpix)
              nPixels0=nPixels0+nPixels1
              !write(*,*) "L1830 nPixels0=", nPixels0
           endif
        endif
     enddo ! l=
  enddo ! k=
  nPixels=nPixels0
end function CheckConnectedPixels



!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! ComputeBackground
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
subroutine ComputeBackground(inputImage, maskImage, bkgImage, xMin, xMax, yMin, yMax, &
     maskAver, maskMax)
  ! Input and unmodified variables
  integer(int32), intent(in) :: xMin, xMax, yMin, yMax                 
  real(single), intent(in) :: inputImage(xMin:xMax, yMin:yMax)  
  real(single), intent(in) :: maskImage(xMin:xMax, yMin:yMax)
  real(single), intent(inout) :: bkgImage(xMin:xMax, yMin:yMax)     
  real(single), intent(out) :: maskAver
  real(single), intent(out) :: maskMax
  ! Local variables
  integer(int32) :: i, j, k,l
  integer(int32) :: dim(2)
!  real(single) :: sum1
  real(single) :: sum2
  integer(int32) :: ismoothWindow, maskWindow
  real(single) :: threshold
  real(single), allocatable, dimension(:) :: vect
  integer(int32) :: vectSize!, nElem
  integer(int32) :: allocErr   
  integer           :: istat
  real(single), allocatable, dimension(:, :) :: tmpImage

!---------------------------------------------
 
  ismoothWindow=4
  vectSize=ismoothWindow*16+4

  allocate (tmpImage(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "tmpImage")  

  allocate (vect(0:vectSize-1), stat = allocErr)
  call CheckAllocation(allocErr, "vect")  

  dim(1) = lbound(inputImage,1)
  dim(2)= ubound(inputImage,1)

  dim(1) = lbound(inputImage,2)
  dim(2)= ubound(inputImage,2)


  !tmpImage=inputImage

! find average and maximal values of the mask
  maskAver=0.0
  maskMax=0.0
  do i=xMin,xMax
     do j=yMin,yMax
        if (maskImage(i,j) .gt. 0.0) then
           maskAver=maskAver+maskImage(i,j)
           sum2=sum2+1.0
           if (maskImage(i,j) .gt. maskMax) maskMax=maskImage(i,j)
        endif
     enddo
  enddo
  if (sum2 .gt. 1.0) maskAver=maskAver/sum2
  threshold=maskAver+0.3*(maskMax-maskAver)


  ! simple smoothing of the input image
  call SimpleImageSmoothing(inputImage, tmpImage, xMin, xMax, yMin, yMax, ismoothWindow)
 
! applying the mask to the smoothed image
!  maskWindow=ismoothWindow/2
  maskWindow=ismoothWindow*8
  do i=xMin,xMax
     do j=yMin,yMax!!
        if (maskImage(i,j) .gt. 0.0) then
           !if (maskImage(i,j) .lt. maskAver) then
              do k=i-maskWindow+2,i+maskWindow+2
                 do l=j-maskWindow+2,j+maskWindow+2
                    if (k .ge. xMin .and. k .le. xMax) then
                       if (l .ge. yMin .and. l .le. yMax) then
                          tmpImage(k,l)=0.0
                       endif
                    endif
                 enddo
              enddo
           !else
           !   !if (maskImage(i,j) .lt. maskAver+0.5*maskMax) then
           !      do k=i-ismoothWindow+2,i+ismoothWindow+2
           !         do l=j-ismoothWindow+2,j+ismoothWindow+2
           !            if (k .ge. xMin .and. k .le. xMax) then
           !               if (l .ge. yMin .and. l .le. yMax) then
           !                  tmpImage(k,l)=0.0
           !               endif
           !            endif
           !         enddo
           !      enddo
           !   !endif
           !endif
        endif
     enddo
  enddo

! median smoothing of the masked image
  call MedianImageSmoothing(tmpImage, bkgImage, xMin,xMax, yMin, yMax, ismoothWindow)
  call MedianImageSmoothing(bkgImage, tmpImage, xMin,xMax, yMin, yMax, ismoothWindow)
  call MedianImageSmoothing(tmpImage, bkgImage, xMin,xMax, yMin, yMax, ismoothWindow)
  call MedianImageSmoothing(bkgImage, tmpImage, xMin,xMax, yMin, yMax, ismoothWindow)
  call SimpleImageSmoothing(tmpImage, bkgImage, xMin, xMax, yMin, yMax, ismoothWindow)


  if(allocated(vect)) then
     deallocate(vect,stat=istat)
     call CheckDeallocation(allocErr, "vect")
  end if
  if(allocated(tmpImage)) then
     deallocate(tmpImage,stat=istat)
     call CheckDeallocation(allocErr, "tmpImage")
  end if

end subroutine ComputeBackground


!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! SimpleImageSmoothing
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
subroutine SimpleImageSmoothing(inputImage, outputImage, xMin, xMax, yMin, yMax, ismoothWindow)
    ! Input and unmodified variables
  integer(int32), intent(in) :: xMin, xMax, yMin, yMax                 
  real(single), intent(in) :: inputImage(xMin:xMax, yMin:yMax)  
  real(single), intent(inout) :: outputImage(xMin:xMax, yMin:yMax)
  integer(int32), intent(in) :: ismoothWindow
     
  ! Local variables
  integer(int32) :: i, j, k,l
  real(single) :: sum1,sum2

!---------------------------------------------
  ! simple smoothing of the input image
  do i=xMin,xMax
     do j=yMin,yMax!!
        sum1=0.0
        sum2=0.0
        do k=i-ismoothWindow,i+ismoothWindow
           do l=j-ismoothWindow,j+ismoothWindow
              if (k .ge. xMin .and. k .le. xMax) then
                 if (l .ge. yMin .and. l .le. yMax) then
                    sum1=sum1+inputImage(k,l)
                    sum2=sum2+1.0
                 endif
              endif
           enddo
        enddo
        if (sum2 .gt. 1.0) sum1=sum1/sum2
        outputImage(i,j)=sum1
     enddo
  enddo
end subroutine SimpleImageSmoothing

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! SimpleImageSmoothing
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
subroutine SimpleImageSmoothing2(inputImage, outputImage, xMin, xMax, yMin, yMax, ismoothWindow)
    ! Input and unmodified variables
  integer(int32), intent(in) :: xMin, xMax, yMin, yMax                 
  real(single), intent(in) :: inputImage(xMin:xMax, yMin:yMax)  
  real(single), intent(inout) :: outputImage(xMin:xMax, yMin:yMax)
  integer(int32), intent(in) :: ismoothWindow
     
  ! Local variables
  integer(int32) :: i, j, k,l
  real(single) :: sum1,sum2

!---------------------------------------------
  ! simple smoothing of the input image
  do i=xMin,xMax
     do j=yMin,yMax!!
        sum1=0.0
        sum2=0.0
        ! cmoothing only for the positive pixels
        if (inputImage(i,j) .gt. 0.0) then
           do k=i-ismoothWindow,i+ismoothWindow
              do l=j-ismoothWindow,j+ismoothWindow
                 if (k .ge. xMin .and. k .le. xMax) then
                    if (l .ge. yMin .and. l .le. yMax) then
                       if (inputImage(k,l) .gt. 0.0) then
                          sum1=sum1+inputImage(k,l)
                          sum2=sum2+1.0
                       endif
                    endif
                 endif
              enddo
           enddo
           if (sum2 .gt. 1.0) sum1=sum1/sum2
           outputImage(i,j)=sum1
        endif
     enddo
  enddo
end subroutine SimpleImageSmoothing2


!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! MedianImageSmoothing
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
subroutine MedianImageSmoothing(inputImage, outputImage, xMin, xMax, yMin, yMax, ismoothWindow)
    ! Input and unmodified variables
  integer(int32), intent(in) :: xMin, xMax, yMin, yMax                 
  real(single), intent(in) :: inputImage(xMin:xMax, yMin:yMax)  
  real(single), intent(inout) :: outputImage(xMin:xMax, yMin:yMax)
  integer(int32), intent(in) :: ismoothWindow
  real(single), allocatable, dimension(:) :: vect
  integer(int32) :: vectSize, nElem
  integer(int32) :: allocErr   
  integer           :: istat
  integer(int32) :: i, j, k,l
  real(single) :: sum1

!---------------------------------------------
  vectSize=ismoothWindow*16+4

  allocate (vect(0:vectSize-1), stat = allocErr)
  call CheckAllocation(allocErr, "vect")  

! median smoothing of the masked image
  outputImage=0.0
  do i=xMin, xmax
     do j=yMin,yMax
        ! fill in the vector (cross-shaped aperture)
        vect=0.0
        nElem=0
        l=j
        do k=i-ismoothWindow*2,i+ismoothWindow*2
           if (k .ge. xMin .and. k .le. xMax) then
              if (inputImage(k,l) .gt. 0.0) then
                 vect(nElem)=inputImage(k,l)
                 nElem=nElem+1
              endif
           endif
        enddo
        k=i
        do l=j-ismoothWindow*2,j+ismoothWindow*2
           if (l .ge. yMin .and. l .le. yMax) then
              if (inputImage(k,l) .gt. 0.0) then
                 vect(nElem)=inputImage(k,l)
                 nElem=nElem+1
              endif
           endif
        enddo
        if (nElem .gt. 0) then
           sum1=smoothMedian(vect,vectSize,nElem)
           outputImage(i,j)=sum1
        endif
     enddo
  enddo

  if(allocated(vect)) then
     deallocate(vect,stat=istat)
     call CheckDeallocation(allocErr, "vect")
  end if

end subroutine MedianImageSmoothing

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! MedianImageSmoothing2
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
subroutine MedianImageSmoothing2(inputImage, outputImage, xMin, xMax, yMin, yMax, ismoothWindow)
    ! Input and unmodified variables
  integer(int32), intent(in) :: xMin, xMax, yMin, yMax                 
  real(single), intent(in) :: inputImage(xMin:xMax, yMin:yMax)  
  real(single), intent(inout) :: outputImage(xMin:xMax, yMin:yMax)
  integer(int32), intent(in) :: ismoothWindow
  real(single), allocatable, dimension(:) :: vect
  integer(int32) :: vectSize, nElem
  integer(int32) :: allocErr   
  integer           :: istat
  integer(int32) :: i, j, k,l
  real(single) :: sum1

!---------------------------------------------
!  vectSize=ismoothWindow*16+4
  vectSize=(ismoothWindow+1)*(ismoothWindow+1)*4

  allocate (vect(0:vectSize-1), stat = allocErr)
  call CheckAllocation(allocErr, "vect")  

! median smoothing of the masked image
  outputImage=-1.0
  do i=xMin, xMax
     do j=yMin,yMax
        if (inputImage(i,j) .gt. 0.0) then 
           ! fill in the vector (cross-shaped aperture)
           vect=0.0
           nElem=0
           l=j
           do k=i-ismoothWindow,i+ismoothWindow
              if (k .ge. xMin .and. k .le. xMax) then
                 do l=j-ismoothWindow,j+ismoothWindow
                    if (l .ge. yMin .and. l .le. yMax) then
                       if (inputImage(k,l) .gt. 0.0) then
                          vect(nElem)=inputImage(k,l)
                          nElem=nElem+1
                       endif
                    endif
                 enddo
              endif
           enddo
           !---------
           if (nElem .gt. 0) then
              sum1=smoothMedian(vect,vectSize,nElem)
              outputImage(i,j)=sum1
           endif
        endif ! inputImage gt 0.0
     enddo
  enddo

  if(allocated(vect)) then
     deallocate(vect,stat=istat)
     call CheckDeallocation(allocErr, "vect")
  end if

end subroutine MedianImageSmoothing2

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! MedianImageSmoothing3
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! smoothing only those pixels that has negative values -3
subroutine MedianImageSmoothing3(inputImage, outputImage, xMin, xMax, yMin, yMax, ismoothWindow)
    ! Input and unmodified variables
  integer(int32), intent(in) :: xMin, xMax, yMin, yMax                 
  real(single), intent(in) :: inputImage(xMin:xMax, yMin:yMax)  
  real(single), intent(inout) :: outputImage(xMin:xMax, yMin:yMax)
  integer(int32), intent(in) :: ismoothWindow
  real(single), allocatable, dimension(:) :: vect
  integer(int32) :: vectSize, nElem
  integer(int32) :: allocErr   
  integer           :: istat
  integer(int32) :: i, j, k,l
  real(single) :: sum1

!---------------------------------------------
!  vectSize=ismoothWindow*16+4
  vectSize=(ismoothWindow+1)*(ismoothWindow+1)*4

  allocate (vect(0:vectSize-1), stat = allocErr)
  call CheckAllocation(allocErr, "vect")  

! median smoothing of the masked image
  outputImage=0.0
  do i=xMin, xMax
     do j=yMin,yMax
        ! fill in the vector (cross-shaped aperture)
        if (inputImage(i,j) .eq. -3.0) then
           vect=0.0
           nElem=0
           l=j
           do k=i-ismoothWindow,i+ismoothWindow
              if (k .ge. xMin .and. k .le. xMax) then
                 do l=j-ismoothWindow,j+ismoothWindow
                    if (l .ge. yMin .and. l .le. yMax) then
                       if (inputImage(k,l) .gt. 0.0) then
                          vect(nElem)=inputImage(k,l)
                          nElem=nElem+1
                       endif
                    endif
                 enddo
              endif
           enddo
           !---------
           if (nElem .gt. 0) then
              sum1=smoothMedian(vect,vectSize,nElem)
              outputImage(i,j)=sum1
           endif
        else
           outputImage(i,j)=inputImage(i,j)
        endif
     enddo
  enddo

  if(allocated(vect)) then
     deallocate(vect,stat=istat)
     call CheckDeallocation(allocErr, "vect")
  end if

end subroutine MedianImageSmoothing3

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! Median smoothing of a vector
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
real function smoothMedian(vect, vectSize, nmax)
  integer(int32), intent(in) :: vectSize
  integer(int32), intent(in) :: nmax ! number of significant elements in the vector, including 0-th
  real(single), intent(in) :: vect(0:vectSize-1)
  real(kind=double) :: a(0:vectSize-1) ! auxiliar vector
  integer :: k,l
  real(single) :: dum
  if (nmax .gt. 2) then
     ! Copy the vector
     a=vect
     ! ordering the vector a
     do k=0,nmax-2
        do l=k+1, nmax-1
           if (a(l) .lt. a(k)) then
              dum=a(k)
              a(k)=a(l)
              a(l)=dum
           endif
        enddo
     enddo
     if (mod(nmax,2) .eq. 0) then
        l=nmax/2
        smoothMedian=(a(l-1)+a(l))/2.0
     else
        l=(nmax-2)/2
        smoothMedian=a(l)
     endif
  else
     smoothMedian=0.0
     do k=0,nmax-1
        smoothMedian=smoothMedian+vect(k)
        dum=dum+1.0
     enddo
     if (dum .gt. 1.0) smoothMedian=smoothMedian/dum
  endif
end function smoothMedian

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
subroutine CircleMaskSmoothing(inputImage, outputImage, xMin, xMax, yMin, yMax, ismoothWindow)
    ! Input and unmodified variables
  integer(int32), intent(in) :: xMin, xMax, yMin, yMax                 
  integer(int32), intent(in) :: inputImage(xMin:xMax, yMin:yMax)  
  integer(int32), intent(inout) :: outputImage(xMin:xMax, yMin:yMax)
  integer(int32), intent(in) :: ismoothWindow
     
  ! Local variables
  integer(int32) :: i, j, k,l
  real(single) :: sum1,sum2!, sum3
  real(single) :: distance!, meanValue

!---------------------------------------------
  ! simple smoothing of the input image
  do i=xMin,xMax
     do j=yMin,yMax!!
        sum1=0.0
        sum2=0.0

        do k=i-ismoothWindow,i+ismoothWindow
           do l=j-ismoothWindow,j+ismoothWindow
              if (k .ge. xMin .and. k .le. xMax) then
                 if (l .ge. yMin .and. l .le. yMax) then
                    distance=sqrt(real((i-k)*(i-k)+(j-l)*(j-l)))
                    if (distance .lt. ismoothWindow) then
                       sum1=sum1+inputImage(k,l)
                       sum2=sum2+1.0
                    endif
                 endif
              endif
           enddo
        enddo
        if (sum1 .gt. 10.0) then
           do k=i-2,i+2
              do l=j-2,j+2
                 if (k .ge. xMin .and. k .le. xMax) then
                    if (l .ge. yMin .and. l .le. yMax) then
                       outputImage(k,l)=1       
                    endif
                 endif
              enddo
           enddo
        endif
     enddo
  enddo
end subroutine CircleMaskSmoothing




!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! FindCentralEnhancement
! VNY 28.08.2008
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

subroutine FindCentralEnhancement(image, outBkgImage, xMin, xMax, yMin, yMax, ismoothWindow)
  ! Input and unmodified variables
!  integer(int32), intent(in) :: crowdedField
  integer(int32), intent(in) :: xMin, xMax, yMin, yMax                 
  real(single), intent(inout) :: image(xMin:xMax, yMin:yMax)
  real(single), intent(inout) :: outBkgImage(xMin:xMax, yMin:yMax)
!  real(single), intent(in) :: inputMaskImage(xMin:xMax, yMin:yMax)
!  real(single), intent(in) :: omgXmin, omgYmin
!  real(kind=double), intent(in) :: grismAngle
  real(kind=double) :: rollangle
!  integer(kind=int32), intent(in) :: windowX0, windowX1, windowY0, windowY1
  ! Local variables
!  integer(int32) :: i, j, k,l, i1,j1
  integer :: i, j, k,l, i1,j1
  integer(int32), intent(in) :: ismoothWindow
  integer(int32) :: ismoothWindow2, ismoothwindow3
  real(single), allocatable, dimension(:, :) :: bkgImage
  real(single), allocatable, dimension(:, :), target :: bkgSubImage
  real(single), allocatable, dimension(:, :), target :: bkgSubImage2
  real(single), allocatable, dimension(:, :), target :: safeImage
  real(single), allocatable, dimension(:, :) :: flatImage
  real(single), allocatable, dimension(:, :) :: flatImage2
  real(single), allocatable, dimension(:, :) :: maskImage
  integer(int32), allocatable, dimension(:, :) :: imaskImage
  integer(int32), allocatable, dimension(:, :) :: imaskImage2
  real(single), allocatable, dimension(:, :) :: imaskImage3
!  integer(int32), intent(in) :: nSourcesMax
!  real(single), intent(out) :: xPos(1:nSourcesMax)
!  real(single), intent(out) :: yPos(1:nSourcesMax)
!  real(single), intent(out) :: semiA(nSourcesMax)
!  real(single), intent(out) :: semiB(nSourcesMax)
!  real(single), intent(out) :: theta(nSourcesMax)
  integer(int32) :: nSources
!  real(single), intent(in) :: nSigma

! 16.12.2004
  !integer(int32), intent(out) :: nSourcesList
  !type(SourceLinkedList), intent(inout) :: sourceList 
  !type(SourceParameters) :: srcParametersData

  real(single) :: sum1,sum2,sum3
  real(single) :: background
  integer(int32) :: allocErr   
  integer           :: istat
  integer(int32) :: dim(2)
  integer(int32) :: idoubleWindow
  real(single), dimension(-ismoothWindow:ismoothWindow,-ismoothWindow:ismoothWindow ) :: template1
  real(single), dimension(-ismoothWindow:ismoothWindow,-2*ismoothWindow-1:2*ismoothWindow+1) :: template2
  real(single) :: umbral
  real(single) :: r11,r12,r21,r22
  integer(int32) :: nPixels, nPixels0
  character(len=1024) :: regionFileName
  real(single) :: sumX, sumY
  real(single) :: sumValue, bkgValue
!  real(single) :: x0,y0
!  real(single) ::  x1, y1
  integer(int32) :: flag
!  integer(int32) ::radius
  integer(int32) :: reject
!  real(single) :: sqrMax, sqrMin, sqrAver
!  integer(int32) :: imaskValue
!  real(single) :: threshold, maskAver, maskMax
  real(single) :: sumValue2a, bkgValue2a!, bkgValue3a
!  integer(int32) :: ima,jma

!-- 13.01.05
  real(kind=single), pointer :: oswImage(:, :)
  real(kind=single), pointer :: bsoswImage(:, :)
  integer(kind=int16) :: sourceNumber
  integer(kind=int32) :: iMin, iMax, jMin, jMax
  real(kind=single) :: xPos, yPos, xPosErr, yPosErr, aValue, bValue, theta 
!--

  integer(int32) :: nSourcesMax, verbFlag
  real(single) :: rMax, rMin, rAver, rAver2, rnAver, bkgFactor, bkgFactor2
  integer(int32) :: fillCrit, fillingFactor
  integer(int32) :: currentNpix
  real(single) :: checkBSimage, checkBSimage2
  integer(int32) :: ibs, jbs
!  real(kind=double)  :: sinRollAngle, cosRollAngle
!  real(double) ::  xposcorr,  yposcorr
!  real(kind=double)  :: trans(6), itrans(6) ! rotation transformation matrix
  logical :: flagl, flagr, flagu, flagd
  real(single) :: vl,vr

  ismoothWindow2=6
  ismoothWindow3=6

  bkgValue=1.0
  bkgFactor=2.0
  bkgFactor2=1.2
  nSourcesMax=20000

  regionFileName="sources.reg"

  template1=0.0
  do i=-ismoothWindow, ismoothWindow
     do j=ismoothWindow-1,ismoothWindow
        k=i*j
        if (k .lt. 0) template1(i,j)=-1.0
        if (k .gt. 0) template1(i,j)=1.0
     enddo
  enddo

  template2=0.0
  do j=ismoothWindow*2-1,ismoothWindow*2+1
     template2(:,j)=1.0
     template2(:,-j)=-1.0
  enddo

  do j=-2,2
     template2(:,j)=1.0
  enddo

  idoubleWindow=ismoothWindow*2+1


  dim(1) = lbound(image,1)
  dim(2)= ubound(image,1)

  dim(1) = lbound(image,2)
  dim(2)= ubound(image,2)

  allocate (bkgImage(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "bkgImage")  

  allocate (bkgSubImage(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "bkgSubImage")  

  allocate (flatImage(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "flatImage")  

  allocate (flatImage2(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "flatImage2")  

  allocate (bkgSubImage2(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "bkgSubImage2")  

  allocate (safeImage(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "safeImage")  

  allocate (maskImage(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "maskImage")  

  allocate (imaskImage(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "maskImage")  

  allocate (imaskImage2(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "maskImage2")  

  allocate (imaskImage3(xMin : xMax, yMin : yMax), stat = allocErr)
  call CheckAllocation(allocErr, "maskImage3")  

  safeImage=image
  oswImage => safeImage


  !---- 08.09.2008 ----
  ! shift the image by 3 pixels vertical
  !-------------------------------------
  call MedianImageSmoothing2(image, bkgImage, xMin,xMax, yMin, yMax, ismoothWindow2)
  !bkgImage=image
  do i=xMin, xMax
     do j=yMin,yMax-3
        image(i,j)=image(i,j)-bkgImage(i,j+3)
     enddo
  enddo
  !imaskImage3=image
  !---- 08.09.2008 ----^^^^
  
  ! shift the image by 3 pixels horizontal

  !call MedianImageSmoothing2(image, bkgImage, xMin,xMax, yMin, yMax, ismoothWindow2)
  !call SimpleImageSmoothing2(bkgImage, image, xMin, xMax, yMin, yMax, ismoothWindow3)
  !call MedianImageSmoothing2(bkgImage, image, xMin,xMax, yMin, yMax, ismoothWindow2)
  !imaskImage3=image
  bkgImage=image
  !image=bkgImage


  do i=xMin, xMax-3
     do j=yMin,yMax
        if (image(i,j) .gt. 0) then
           image(i,j)=image(i,j)-bkgImage(i+3,j)
        endif
     enddo
  enddo

  !call MedianImageSmoothing2(image, imaskImage3, xMin,xMax, yMin, yMax, ismoothWindow2)
  ! imaskImage3 is stored to check it at the end of the programme
  ! NB
  !imaskImage3=image
  ! NB

  !---- 08.09.2008 ----
  !maskImage=0.0
  !umbral=3.0 !10.0
  !do i=xMin,xMax
  !   do j=yMin,yMax!!
  !      if (i-2 .ge. xMin .and. i+2 .le. xMax) then
  !         if (j-2 .ge. yMin .and. j+2 .le. yMax) then
  !            r11=image(i-2,j-2)
  !            r12=image(i+1,j-2)
  !            r21=image(i-2,j+1)
  !            r22=image(i+1,j+1)
  !            if (r21 .lt. -umbral &
  !                 .and. r11 .gt. umbral &
  !                 .and. r22 .gt. umbral &
  !                 .and. r12 .lt. -umbral) then
  !               maskImage(i,j)=maskImage(i,j)+r11-r12-r21+r22
  !            endif
  !         endif
  !      endif!
  !   enddo
  !enddo
  !imaskImage3=maskImage

  !---- 08.09.2008 ----^^^^

!  !************************
!  ! compute the flat image
!  !************************
!  flatImage=0.0
!  flatImage2=0.0
!  where (safeImage .gt. 0.0) flatImage=1.0
!
!  ! remove the standing alone pixels
!  do i=xMin+2,xMax-2
!     do j=yMin+2,yMax-2
!        sum1=0.0
!        do i1=i-2,i+2
!           do j1=j-2,j+2
!              sum1=sum1+flatImage(i1,j1)
!           enddo
!        enddo
!        if (sum1 .gt. 14.0) then
!           flatImage2(i,j)=bkgValue
!        endif
!     enddo
!  enddo
!
!  !shift the image by 3 pixels horizontal
!  flatImage=flatImage2
!
!
!  do i=xMin, xMax-3
!     do j=yMin,yMax
!        flatImage2(i,j)=flatImage2(i,j)-flatImage(i+3,j)
!     enddo
!  enddo
!
!  
!  ! remove the borders using the above shifted flat image
!  !do i=xMin+3, xMax-3
!  !   do j=yMin+3,yMax-3
!  !      if (flatImage2(i,j) .ne. 0.0) then
!  !         do i1=i-3,i+3
!  !            do j1=j-3,j+3
!  !               image(i1,j1)=0.0
!  !            enddo
!  !         enddo
!  !      endif
!  !   enddo
!  !enddo

  imaskImage2=image
!  imaskImage3=image

  ! calculate the minimum and maximum of the signal
  sum1=999999.0 ! minimum
  sum2=-999999.0 ! maximum
  do j=yMin+3,yMax-3
     do i=xMin+3,xMax-3
        if (safeImage(i,j) .gt. 0.0) then
           sum3=image(i+1,j)-image(i-1,j)
           if (sum3 .gt. sum2) sum2=sum3
           if (sum3 .lt. sum1) sum1=sum3
        endif
     enddo
  enddo


  ! calculate the average of the signal
  sum1=0.0
  sum2=0.0
  do j=yMin+3,yMax-3
     do i=xMin+3,xMax-3
        if (safeImage(i,j) .gt. 0.0) then
           sum3=image(i+1,j)-image(i-1,j)
           if (sum3 .gt. 0.0) then
              sum1=sum1+sum3!abs(image(i,j))
              sum2=sum2+1.0
           endif
        endif
     enddo
  enddo
  if (sum2 .gt. 1.0) sum2=sum1/sum2


  ! smooth the image in the horizontal direction
  do j=yMin+3,yMax-3
     do i=xMin+3,xMax-3
        if (image(i+1,j) .gt. 0.0 .and. image(i-1,j) .lt. 0.0) then 
           sum1=image(i+1,j)-image(i-1,j)
           if (sum1 .gt. sum2) then
              do i1=i-4,i+8
                 if (i1 .ge. xMin .and. i1 .le. xMax) then               
                    imaskImage2(i1,j)=50.0
                 endif
              enddo
           endif
        endif
     enddo
  enddo
  image=imaskImage2
  !imaskImage3=image

  ! remove the negative values of the shifted image
  do i=xMin+3, xMax-3
     do j=yMin+3,yMax-3
        if (abs(image(i,j)) .gt. 40) then
           image(i,j)=-50.0
        endif
     enddo
  enddo
  

  ! NB
  imaskImage2=safeImage
  where (image .lt. -40.0) imaskImage2=-2.0
  ! NB
  image=imaskImage2
  !imaskImage3=image


  ! smooth the image to get the background without sources
  !do i=xMin, xMax
  !   do j=yMin,yMax
  !      ! Calculate average only for the empty pixels
  !      if ((image(i,j)) .eq. -2.0) then
  !         sum1=0.0
  !         sum2=0.0
  !         do i1=-25,25
  !            do j1=-25,25
  !               if (i+i1 .gt. xMin .and. i+i1 .lt. xMax) then
  !                  if (j+j1 .gt. yMin .and. j+j1 .lt. yMax) then
  !                     if (imaskImage2(i+i1,j+j1) .gt. 0.0) then
  !                        sum1=sum1+imaskImage2(i+i1,j+j1)
  !                        sum2=sum2+1.0
  !                     endif
  !                  endif
  !               endif
  !            enddo ! j1=-50,50
  !         enddo ! i1=-50,50
  !         if (sum2 .gt. 1.0) sum1=sum1/sum2
  !         image(i,j)=sum1
  !      endif
  !  enddo
  !enddo
!-----
  !-----------------------------------------------------------
  ! smooth the image to get the background without sources 222
  ! sommthing along the horizontal direction
  !----------------------------------------------------------
  do i=xMin, xMax
     do j=yMin,yMax
        !flag=0
        !if (i .eq. 627 .and. j .eq. 542) flag=1
        ! Calculate average only for the empty pixels
        if ((image(i,j)) .eq. -2.0) then
           vl=-2.0 ! value left
           do i1=1,25
              if (i-i1 .gt. xMin) then
                 !if (flag .eq. 1) then
                 !   write(*,*) "L2048 left i1=", i1, " image=", imaskImage2(i-i1,j)
                 !endif
                 if (imaskImage2(i-i1,j) .gt. 0.0) then
                    vl=imaskImage2(i-i1,j)
                    exit
                 endif
              else
                 vl=-2.0
                 exit
              endif
           enddo
           vr=-2.0 ! value right
           do i1=1,25
              if (i+i1 .lt. xMax) then
                 !if (flag .eq. 1) then
                 !   write(*,*) "L2063 right i1=", i1, " image=", imaskImage2(i-i1,j)
                 !endif
                 if (imaskImage2(i+i1,j) .gt. 0.0) then
                    vr=imaskImage2(i+i1,j)
                    exit
                 endif
              else
                 vr=-2.0
                 exit
              endif
           enddo
           !if (flag .eq. 1) then
           !   write(*,*) "L2075 vl=", vl, " vr=", vr
           !endif
           if (vl .gt. 0.0 .and. vr .gt. 0.0) then
              vl=(vl+vr)/2.0
           else
              if (vl .le. 0.0 .and. vr .gt. 0.0) then
                 vl=vr
              endif
           endif
           image(i,j)=vl
        endif
     enddo
  enddo

!-----------------------------------------------------------------------
! Remove remaining empty pixels by smoothing along the vertical direction
! (repeat the above smoothing but in the vertical direction)
!----------------------------------------------------------------------
  imaskImage2=image
  do i=xMin, xMax
     do j=yMin,yMax
        ! Calculate average only for the empty pixels
        if ((image(i,j)) .eq. -2.0) then
           vl=-2.0 ! value left (lower)
           do i1=1,25
              if (j-i1 .gt. yMin) then
                 if (imaskImage2(i,j-i1) .gt. 0.0) then
                    vl=imaskImage2(i,j-i1)
                    exit
                 endif
              else
                 vl=-2.0
                 exit
              endif
           enddo
           vr=-2.0 ! value right (upper)
           do i1=1,25
              if (j+i1 .lt. yMax) then
                 if (imaskImage2(i,j+i1) .gt. 0.0) then
                    vr=imaskImage2(i,j+i1)
                    exit
                 endif
              else
                 vr=-2.0
                 exit
              endif
           enddo
           if (vl .gt. 0.0 .and. vr .gt. 0.0) then
              vl=(vl+vr)/2.0
           else
              if (vl .le. 0.0 .and. vr .gt. 0.0) then
                 vl=vr
              endif
           endif
           image(i,j)=vl
        endif
     enddo
  enddo

!--------------------------------

  call MedianImageSmoothing2(image, imaskImage3, xMin,xMax, yMin, yMax, ismoothWindow2)
  !imaskImage3=image
  outBkgImage=imaskImage3

  ! find max and min of the background
  sum1=99999.0 ! min
  sum3=-99999.0 ! max
  do j=yMin+2,yMax-2
     do i=xMin+2,xMax-2
           if (imaskImage3(i,j) .lt. sum1) sum1=imaskImage3(i,j)
           if (imaskImage3(i,j) .gt. sum3) sum3=imaskImage3(i,j)
     enddo
  enddo

 ! write(*,*) "L1125 sum1=", sum1, " (min);   sum3=", sum3, " (max)"

  ! calculate the average of the background
  sum1=0.0
  sum3=0.0
  do j=yMin+2,yMax-2
     do i=xMin+2,xMax-2
        if (imaskImage3(i,j) .gt. 0.0) then
           sum1=sum1+imaskImage3(i,j)
           sum3=sum3+1.0
        endif
     enddo
  enddo
  if (sum3 .gt. 1.0) background=sum1/sum3
!  write(*,*) "L1139 background=", background

  ! remove the background without smoothing
  !imaskImage3=-1.0
  !do i=xMin, xMax
  !   do j=yMin,yMax
  !      if (image(i,j) .gt. 0.0) then
  !         imaskImage3(i,j)=safeImage(i,j) - image(i,j)
  !      !else
  !      !   imaskImage3(i,j)=safeImage(i,j)
  !      endif
  !   enddo
  !enddo

  !call MedianImageSmoothing2(image, imaskImage3, xMin,xMax, yMin, yMax, ismoothWindow2)
  !call SimpleImageSmoothing(imaskImage3, image, xMin, xMax, yMin, yMax, ismoothWindow)


  !call SimpleImageSmoothing2(image, imaskImage3, xMin, xMax, yMin, yMax, ismoothWindow)

  !imaskImage3=image

!  ! calculate the average of the smoothed background imaskImage3
!  sum1=0.0
!  sum3=0.0
!  do j=yMin+2,yMax-2
!     do i=xMin+2,xMax-2
!        if (abs(imaskImage3(i,j)) .gt. 0.0) then
!           sum1=sum1+imaskImage3(i,j)
!           sum3=sum3+1.0
!        endif
!     enddo
!  enddo
!  if (sum3 .gt. 1.0) sum3=sum1/sum3
!  write(*,*) "L1950 sum3=", sum3

  ! Remove the smoothed background from the original image
!  imaskImage3=safeImage - image
  do i=xMin, xMax
     do j=yMin,yMax
        if (imaskImage3(i,j) .gt. 0.0) then
           image(i,j)=safeImage(i,j) - imaskImage3(i,j)
           !image(i,j)=image(i,j)+background
        else
           image(i,j)=safeImage(i,j)
        endif
     enddo
  enddo

  ! Add the background
  do i=xMin, xMax
     do j=yMin,yMax
        if (imaskImage3(i,j) .gt. 0.0) then
           image(i,j)=image(i,j)+background
        else
           image(i,j)=-1.0
        endif
     enddo
  enddo

  ! Remove stand-alone negative pixels
  !do i=xMin+3, xMax-3
  !   do j=yMin+3,yMax-3
  !      if (image(i,j) .lt. 0.0) then
  !         flagl=.false.
  !         flagr=.false.
  !         flagu=.false.
  !         flagd=.false.
  !         do j1=-1,1
  !            if (image(i-1,j+j1) .gt. 0.0) flagl=.true.
  !            if (image(i+1,j+j1) .gt. 0.0) flagr=.true.
  !         enddo
  !         do j1=-2,2
  !            if (image(i-2,j+j1) .gt. 0.0) flagl=.true.
  !            if (image(i+2,j+j1) .gt. 0.0) flagr=.true.
  !         enddo
  !         do j1=-3,3
  !            if (image(i-3,j+j1) .gt. 0.0) flagl=.true.
  !            if (image(i+3,j+j1) .gt. 0.0) flagr=.true.
  !         enddo!
!
  !         do i1=-1,1
  !            if (image(i+i1,j+1) .gt. 0.0) flagu=.true.
  !            if (image(i+i1,j-1) .gt. 0.0) flagd=.true.
  !         enddo
  !         do i1=-2,2
  !            if (image(i+i1,j+2) .gt. 0.0) flagu=.true.
  !            if (image(i+i1,j-2) .gt. 0.0) flagd=.true.
  !         enddo
  !         do i1=-3,3
  !            if (image(i+i1,j+3) .gt. 0.0) flagu=.true.
  !            if (image(i+i1,j-3) .gt. 0.0) flagd=.true.
  !         enddo
  !         if (flagl .or. flagr .or. flagu .or. flagd) then
  !            image(i,j)=-3
  !         endif
  !      endif
  !   enddo
  !enddo

  imaskImage3=image 



  image=imaskImage3


!*************************************************
  if(allocated(bkgImage)) then
     deallocate(bkgImage,stat=istat)
     call CheckDeallocation(allocErr, "bkgImage")
  end if

  if(allocated(bkgSubImage)) then
     deallocate(bkgSubImage,stat=istat)
     call CheckDeallocation(allocErr, "bkgSubImage")
  end if

  if(allocated(flatImage)) then
     deallocate(flatImage,stat=istat)
     call CheckDeallocation(allocErr, "flatImage")
  end if

  if(allocated(flatImage2)) then
     deallocate(flatImage2,stat=istat)
     call CheckDeallocation(allocErr, "flatImage2")
  end if

  if(allocated(bkgSubImage2)) then
     deallocate(bkgSubImage2,stat=istat)
     call CheckDeallocation(allocErr, "bkgSubImage2")
  end if

  if(allocated(imaskImage)) then
     deallocate(imaskImage,stat=istat)
     call CheckDeallocation(allocErr, "imaskImage")
  end if

  if(allocated(imaskImage2)) then
     deallocate(imaskImage2,stat=istat)
     call CheckDeallocation(allocErr, "imaskImage2")
  end if

  if(allocated(imaskImage3)) then
     deallocate(imaskImage3,stat=istat)
     call CheckDeallocation(allocErr, "imaskImage3")
  end if

  if(allocated(maskImage)) then
     deallocate(maskImage,stat=istat)
     call CheckDeallocation(allocErr, "maskImage")
  end if

  if(allocated(safeImage)) then
     deallocate(safeImage,stat=istat)
     call CheckDeallocation(allocErr, "safeImage")
  end if

end subroutine FindCentralEnhancement



end module background_mod

